/****************************************************************************\
*                                                                            *
* Function: Palette entrypoints                                              *
*                                                                            *
*       Set, get, and update palette colors.                                 *
*                                                                            *
* Author:                                                                    *
*                                                                            *
*       Dave Schmenk                                                         *
*                                                                            *
* History:                                                                   *
*                                                                            *
*       05/11/93 Dave Schmenk - wrote it.                                    *
*                                                                            *
******************************************************************************
*                                                                            *
*                     Copyright 1992 pellucid, inc.                          *
*                                                                            *
\****************************************************************************/

#include "common.h"
#include "palette.h"

#include "iv1.h"

#define AGR 1
#define WIN_DRIVER 1

#include "invent.h"
#include "gfxboard.h"
#include "gecmds.h"
#include "gr1_dev.h"
#include "iv_ucode.h"


#include "externs.h"

#include "atsubr.h"



                                //
DWORD devpStatus = 0;           //  Global Variable - Used for Windows statuses
                                //

extern int ivhw_state;

//
// This is the system palette defined as IRGB, where I is the index into the
// device palette.
//


#define IRGB_DWORD(i, r, g, b) (((DWORD)(r)) | ((DWORD)(g) << 8) | ((DWORD)(b) << 16) | ((DWORD)(i) << 24))

DWORD devSystemPalette[20] =
{
    IRGB_DWORD(0,   0x00, 0x00, 0x00),
    IRGB_DWORD(1,   0x80, 0x00, 0x00),
    IRGB_DWORD(2,   0x00, 0x80, 0x00),
    IRGB_DWORD(3,   0x80, 0x80, 0x00),
    IRGB_DWORD(4,   0x00, 0x00, 0x80),
    IRGB_DWORD(5,   0x80, 0x00, 0x80),
    IRGB_DWORD(6,   0x00, 0x80, 0x80),
    IRGB_DWORD(7,   0xC0, 0xC0, 0xC0),
    IRGB_DWORD(8,   0xC0, 0xDC, 0xC0),
    IRGB_DWORD(9,   0xA4, 0xC8, 0xF0),
    IRGB_DWORD(246, 0xFF, 0xFB, 0xF0),
    IRGB_DWORD(247, 0xA0, 0xA0, 0xA4),
    IRGB_DWORD(248, 0x80, 0x80, 0x80),
    IRGB_DWORD(249, 0xFF, 0x00, 0x00),
    IRGB_DWORD(250, 0x00, 0xFF, 0x00),
    IRGB_DWORD(251, 0xFF, 0xFF, 0x00),
    IRGB_DWORD(252, 0x00, 0x00, 0xFF),
    IRGB_DWORD(253, 0xFF, 0x00, 0xFF),
    IRGB_DWORD(254, 0x00, 0xFF, 0xFF),
    IRGB_DWORD(255, 0xFF, 0xFF, 0xFF)
};

//
// This table reflects the high intensity colors of the system palette for
// converting color indeces into monochrome white or black.
//

BYTE devColorToMono[20] =
{
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x01,
    0x01,
    0x01,
    0x01,
    0x01,
    0x00,
    0x00,
    0x01,
    0x01,
    0x00,
    0x01,
    0x01,
    0x01
};

//
// Copy of the devices hardware palette.
//

RGBQUAD devPalette[256];

//
// Translation table from logical indeces to device indeces.
//

BYTE devXlateTbl[256];

//
// Translation table from device indeces to logical indeces.
//

BYTE devUnxlateTbl[256];

//
// Translateion table used in Update Colors
//
LPBYTE  devUpdateColorsTbl;

VOID PASCAL devpUpdateColors(SHORT, SHORT, LPBYTE, USHORT, USHORT, USHORT, SHORT, SHORT, SHORT, SHORT, DWORD, SHORT, USHORT);

VOID WINAPI SetPalette
(
    WORD      wStartIndex,
    WORD      wNumEntries,
    LPRGBQUAD lpPalette
)
{
    WORD ii;
    WORD i;

    GRPsetup;
    unsigned short intr_flgs;  /* for saving flags during critcal sections */

    DBGMSG("SetPalette\n\r");


    if (ivhw_state & INV_GR1BIT24) {


        for (i = 0; i < wNumEntries; i+=32) {
            VRWAIT2;    
            intr_flgs = EnterCrit();  /* this should disable interrupts */
            HQMMSB_WR( 1 );
            XMAP_ALOWR((i+wStartIndex) | XMAP_COLSEL);
            XMAP_AHIWR((i+wStartIndex) | XMAP_COLSEL);

            for (ii=0; ((ii<32)&&((i+ii)<wNumEntries)); ii++) {
                XMAP_COLWR(XMAP_RED, lpPalette[i+ii].rgbBlue);
                XMAP_COLWR(XMAP_GREEN,lpPalette[i+ii].rgbGreen);
                XMAP_COLWR(XMAP_BLUE,lpPalette[i+ii].rgbRed);
                XMAP_INCA();
            }
            HQMMSB_WR( 0 );
            LeaveCrit(intr_flgs);  /* this should enable interrupts */
      }

    }
    else {  /* minimum bitplane system */

        for (i = 0; i < wNumEntries; i+=32) {
            VRWAIT2;    
            intr_flgs = EnterCrit();  /* this should disable interrupts */
            HQMMSB_WR( 1 );

            for (ii=0; ((ii<32)&&((i+ii)<wNumEntries)); ii++) {
                DAC_AWR(DACRGB_OFF, i+wStartIndex+ii);

                DAC_COLWR(DACRGB_OFF, lpPalette[i+ii].rgbBlue);
                DAC_COLWR(DACRGB_OFF, lpPalette[i+ii].rgbGreen);
                DAC_COLWR(DACRGB_OFF, lpPalette[i+ii].rgbRed);

             }
             HQMMSB_WR( 0 );
             LeaveCrit(intr_flgs);  /* this should enable interrupts */
        }

    } /* else 8 bit system */

    for (ii = 0; ii < wNumEntries; ii++)
        devPalette[ii + wStartIndex] = lpPalette[ii];

    
}

VOID WINAPI GetPalette
(
    WORD      wStartIndex,
    WORD      wNumEntries,
    LPRGBQUAD lpPalette
)
{
    WORD ii;

    DBGMSG("GetPalette\n\r");

    for (ii = 0; ii < wNumEntries; ii++)
        lpPalette[ii] = devPalette[ii + wStartIndex];
}

VOID WINAPI SetPalTrans
(
    LPWORD lpIndexes
)
{
}

VOID WINAPI GetPalTrans
(
    LPWORD lpIndexes
)
{
}

VOID WINAPI UpdateColors
(
    WORD   wDstX,
    WORD   wDstY,
    WORD   wXext,
    WORD   wYext,
    LPWORD lpTranslate
)
{
}
